/***********************************************************
 *   Bpm Live - Live BPM counting
 *-----------------------------------------------------------
 *  Copyright (c) 2004 Werner Van Belle, see copyright.txt
 *  http://werner.onlinux.be/bpmlive/
 *  e-mail: werner.van.belle@itek.norut.no
 ***********************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <fftw3.h>
#include "bpm-counter.h"
#include "stereo-sample2.h"
#include "bpmlive-signal.h"
#include "memory.h"
#include "bpmlive.h"
#include "bpm-value.h"
#include "cluster.h"
#include "dsp-alsa.h"

void BpmCounter::init(dsp_alsa * d)
{
  /** 
   * We calculate the length of an analysis block. 
   * We aim to include all frequencies starting from
   * at least 50 Hz. If we start at 44100 Hz we must
   * halve the value 11 times. Which means that we 
   * must have at least a buffer of 1<<11
   */
  
  dsp = d;
  int atmost = dsp->get_max_read();
  measured = 1;
  while(measured <= atmost) measured *=2;
  measured/=2;
   
  block = Allocate(measured,sample_type);
  audio.init(measured);
   
  /**
   * Preparation of the fourier windows, for every band we will have
   * a separate plan available. All output is stored in the energy
   * array, which is then used to compute the autocorrelation sequence
   */
  FILE * wisdom_file = fopen(".fftw_wisdow","rb");
  if (wisdom_file)
    {
      printf("Loading fftw wisdow %d\n",fftw_import_wisdom_from_file(wisdom_file));
      fclose(wisdom_file);
    }
  
  en = Allocate(measured * 2,double);
  ts = Allocate(measured * 2,double);
  co = Allocate(measured * 2,double);
  
  printf("Preparing fourier transform (%d)\n",measured);
  forward = fftw_plan_r2r_1d(measured, &(audio[0]), en, FFTW_R2HC, FFTW_HOW);
  backward = fftw_plan_r2r_1d(measured * 2, ts, co, FFTW_HC2R, FFTW_HOW);
  printf("Done\n");
  
  
  /**
   * The fftw wisdom things 
   */
  
  wisdom_file = fopen(".fftw_wisdow","wb");
  if (wisdom_file)
    {
      printf("Saving fftw wisdow\n");
      fftw_export_wisdom_to_file(wisdom_file);
      fclose(wisdom_file);
    }
  
  /**
   * The boundaries
   */	
  
  sample_rate = dsp -> get_wavrate() ;
  double freq = 160.0/60.0;
  b1 = 4 * (int)(sample_rate / freq);
  freq = 90.0/60.0;
  b2 = 4 * (int)(sample_rate / freq);
  
  /**
   * The window mask
   */
  mask = Allocate(measured, double);
  for(int x = 0 ; x < measured ; x ++)
    mask[x] = 0.5 - 0.5 * cos(M_PI*2.0*(double)x/(double)measured);
}

double BpmCounter::measure()
{
  dsp->read(block,measured);
  #ifdef MONO8BIT
  for(int i = 0 ; i < measured ; i ++)
    audio[i]=block[i];
  #endif
  
  #ifdef STEREO16BIT
  for(int i = 0 ; i < measured ; i ++)
    audio[i]=block[i].left+block[i].right;
  #endif
  
  for(int x = 0 ; x < measured ; x ++)
    audio[x]=fabs(audio[x]);
  for(int x = 0 ; x < measured ; x ++)
    audio[x]*=mask[x];
  fftw_execute(forward);
  for(int x = 0 ; x < measured / 2; x ++)
    ts[x]=en[x]=sqrt(en[x]*en[x]+en[measured-x]*en[measured-x]);
  for(int x = measured / 2 ; x < measured * 2 ; x ++)
    ts[x]=0;
  fftw_execute(backward);
  for(int x = 0 ; x < measured ; x ++)
    co[x]/=(measured-x);
  double mf = en[0] = 0;
  int pf = 0;
  for(int x = b1 ;  x < b2 ; x ++)
    {
      int m = 1;
      double E = 0;
      while (m<=32)
	{
	  E += en[measured*m/x];
	       m *= 2;
	}
      E/=6;
      E*=co[x];
      if (E>mf || mf == 0.0)
	{
	  mf = E;
	  pf = x;
	}
    }
  return 4.0 * sample_rate * 60.0 / (double) pf ;
}
