/***********************************************************
 *   Bpm Live - Live BPM counting
 *-----------------------------------------------------------
 *  Copyright (c) 2004 Werner Van Belle, see copyright.txt
 *  http://werner.onlinux.be/bpmlive/
 *  e-mail: werner.van.belle@itek.norut.no
 ***********************************************************/

#include <assert.h>
#include "inout.h"
#include "memory.h"

 /**
  * This is an atypical producer/consumer problem. We don't want
  * to make use of locks because these doesn't really solve a problem.
  * (E.g: what do we do when we have new data which we can't push because
  * there is a lock, busy-wait ? not really performant)
  * As such, we make sure that the buffer has some spare space for new
  * data. The most recent position contains the position of the last written
  * piece of data. As such the most recent block is located at 
  * most_recent_pos-max_read. The total size of the buffer is max_read+
  * max_write.
  *
  * pushing data is done by writing data into the buffer and then positioning
  * the most recent position.
  * pulling data is always done at the most recent position.
  */
template <class T> void Buffer<T>::push(T * source, int length)
{
	assert(length<=max_write);
	// will the block cross the boundary ?
	if (most_recent_pos + length > total_size)
		{
		int chunk  = total_size - most_recent_pos;
		memcpy(content+most_recent_pos,source,chunk*sizeof(T));
		memcpy(content,source+chunk,(length-chunk)*sizeof(T));
		}
	else
		{
		memcpy(content+most_recent_pos,source,length*sizeof(T));
		}
	most_recent_pos = (most_recent_pos + length) % total_size;
}

template <class T> void Buffer<T>::pull(T* target, int length)
{
	assert(length<=max_read);
	// we copy the variable to avoid problems when the 
	// variable is rewritten in the push
	int p = most_recent_pos;
	int s = p - length;
	// will the block cross the boundary ?
	if (s < 0)
		{
		memcpy(target,content+total_size+s,-s*sizeof(T));
		memcpy(target-s,content,(length+s)*sizeof(T));
		}
	else
		{
		memcpy(target,content+s,length*sizeof(T));
		}
}

template <class T> Buffer<T>::Buffer(int mr, int mw)
{
	total_size = mr + mw;
	max_read = mr;
	max_write = mw;
	most_recent_pos = 0;
	content = Allocate(total_size,T);
}
